; MP3 Player, Main Program, http://www.pjrc.com/tech/mp3
; Copyright (c) 2000, PJRC.COM, LLC

; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.

; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

; As a specific exception to the GPL, the executable object code built
; from this source code may be combined with hardware configuration data
; files.  A hardware configuration data file is a set of data that is
; transmitted to an intergrated circuit that is not a general purpose
; microprocessor or microcontroller, in order to establish its normal
; operation.  The process of combining the executable ojbect code built
; from the GPL licensed source code with the hardware configuration data
; shall be considered an aggregation of another work not based on the
; Program.  While the GPL does not restrict use of the program, any
; use restriction associated with the hardware configuration data (for
; example, that it only be used with particular hardware) shall apply
; to the combined file which includes a copy of the hardware configuration
; data.

; Contact: paul@pjrc.com



;*************************************************************
;**                                                         **
;**                  Memory Allocation                      **
;**                                                         **
;*************************************************************


.equ    location, 0x2000        ;where this program will exist


; we'll keep a 4k chunk of DRAM memory from 0000 to 0FFF
.equ	sector_buffer, 0x0E00


; functions built into Paulmon2
.equ    cout, 0x0030            ;Send Acc to serial port
.equ    cin, 0x0032             ;Get Acc from serial port
.equ    phex, 0x0034            ;Print Hex value of Acc
.equ	phex1, 0x002E
.equ    phex16, 0x0036          ;Print Hex value of DPTR
.equ    pstr, 0x0038            ;Print string pointed to by DPTR,
.equ    upper, 0x0040           ;Convert Acc to uppercase
.equ    pint8u, 0x004D          ;print Acc at an integer, 0 to 255
.equ    pint8, 0x0050           ;print Acc at an integer, -128 to 127
.equ    pint16u, 0x0053         ;print DPTR as an integer, 0 to 65535
.equ    newline, 0x0048         ;print CR/LF (13 and 10)


; library functions in bank1 (bank switch special calling)
.equ	call_bank1, 0x0FE0
.equ	xdownload, 0xA040
.equ	sta013_init, 0xA043
.equ	shutdown, 0xA046
.equ	alt_xdownload, 0xA049


; registers implemented in the DRAM, IDE, and DMA controller
.equ	dram_page_cfg,	0xFF00
.equ	ide_data,	0xFF60
.equ	ide_err,	0xFF62
.equ	ide_sec_cnt,	0xFF64
.equ	ide_sector,	0xFF66
.equ	ide_cyl_lsb,	0xFF68
.equ	ide_cyl_msb,	0xFF6A
.equ	ide_head,	0xFF6C
.equ	ide_command,	0xFF6E
.equ	ide_status,	0xFF6E
.equ	ide_control,	0xFF7C
.equ	ide_astatus,	0xFF7E
.equ	ide_rst_bit,	0xFF40
.equ	ide_data_buf_msb, 0xFF43
.equ	dma_ide_dest,	0xFF22
.equ	dma_ide_count,	0xFF24
.equ	irq_dma_ide_ack,0xFF5C
.equ	dma_ide_go,	0xFF58
.equ	irq_ident,	0xFF50
.equ	dma_mp3_src,	0xFF28
.equ	dma_mp3_count,	0xFF2A
.equ	dma_mp3_go,	0xFF59
.equ	irq_dma_mp3_ack,0xFF5D








;IDE Command Constants.  These should never change.
.equ    ide_cmd_recal, 0x10
.equ    ide_cmd_read, 0x20
.equ    ide_cmd_write, 0x30
.equ    ide_cmd_init, 0x91
.equ    ide_cmd_id, 0xEC
.equ    ide_cmd_spindown, 0xE0
.equ    ide_cmd_spinup, 0xE1
.equ	ide_cmd_idle, 0xE1
.equ	ide_cmd_sleep, 0xE6
.equ	ide_cmd_set_features, 0xEF


; internal 8051 memory allocation
.equ	status, 0x08			;1 byte
.equ	sectors_per_cluster, 0x09	;1 byte
.equ	button_state, 0x0A		;1 byte
.equ	state_quit, 0x0B		;1 byte

.equ	lba, 0x0C			;4 bytes
.equ	size, 0x10			;4 bytes
.equ	partition_first_sector, 0x14	;4 bytes
.equ	partition_size, 0x18		;4 bytes
.equ	first_data_sector, 0x1C		;4 bytes

.equ	flags, 0x21
.equ	dir_position_sector, 0x22	;1 byte, 0 to sectors_per_cluster-1
.equ	dir_position_offset, 0x23	;1 byte, 0 to 15
.equ	dir_position_cluster, 0x24	;4 bytes

.equ	root_dir_cluster, 0x28		;4 bytes
.equ	cluster, 0x2C			;4 bytes

.equ	filename, 0x30			;11 bytes
.equ	attrib, 0x3B			;1 byte

.equ	reserved_sector_count, 0x3C	;2 bytes
.equ	tmp_var, 0x3E			;4 bytes
.equ	state_next, 0x42		;1 byte
.equ	state_prev, 0x43		;1 byte
.equ	state_play, 0x44		;1 byte

.equ	stack, 0x50



; internal bit addressable memory allocation
.equ	playing_flag, 0x08
.equ	paused_flag, 0x09
.equ	next_flag, 0x0A
.equ	prev_flag, 0x0B
.equ	end_now_flag, 0x0C





.org    location
.db     0xA5,0xE5,0xE0,0xA5     ;signiture bytes
.db     254,'P',0,0             ;id
.db     0,0,0,0                 ;reserved
.db     0,0,0,0                 ;reserved
.db     0,0,0,0                 ;reserved
.db     0,0,0,0                 ;reserved
.db     0,0,0,0                 ;user defined
.db     255,255,255,255         ;length and checksum (255=unused)
.db     "MP3 Player, Ver 0.5.0",0
.org    location+64             ;executable code begins here



;*************************************************************
;**                                                         **
;**                  Initialize Hardware                    **
;**                                                         **
;*************************************************************


begin:
	mov	sp, #stack
	mov	dptr, #mesg_welcome
	lcall	pstr

	clr	a
	mov	flags, a
	mov	state_quit, a
	mov	state_play, a
	mov	state_next, a
	mov	state_prev, a
	acall	read_buttons
	orl	a, 0x10
	mov	button_state, a

	mov	dptr, #mesg_xilinx_cfg
	lcall	pstr
	mov	dptr, #alt_xdownload
	lcall	call_bank1
	jc	error
	mov	dptr, #mesg_ok
	lcall	pstr

	mov	dptr, #mesg_reset_ide
	lcall	pstr
	acall	ide_hard_reset

	mov	dptr, #mesg_init_mm
	lcall	pstr
	lcall	init_memory_mgr
	jc	error
	clr	a
	mov	r2, a
	mov	r3, a
	lcall	map_block

	mov	dptr, #mesg_sta013_cfg
	lcall	pstr
	mov	dptr, #sta013_init
	lcall	call_bank1
	jc	error
	mov	dptr, #mesg_ok
	lcall	pstr

	mov	dptr, #mesg_ide_init
	lcall	pstr
	acall	ide_init
	lcall	newline
	ajmp	read_mbr

error:
	mov	dptr, #mesg_error
	lcall	pstr
	ajmp	quit







;*************************************************************
;**                                                         **
;**      Detect Partitions and FAT32 Filesystem Info        **
;**                                                         **
;*************************************************************


read_mbr:
	mov	lba+3, #0
	mov	lba+2, #0
	mov	lba+1, #0
	mov	lba+0, #0

	mov	dptr, #mesg_read_mbr
	lcall	pstr
	mov	dptr, #sector_buffer
	acall	ide_read_sector
	mov	dptr, #sector_buffer + 0x1FE
	movx	a, @dptr
	cjne	a, #0x55, bad_mbr
	inc	dptr
	movx	a, @dptr
	cjne	a, #0xAA, bad_mbr

	mov	dptr, #sector_buffer + 0x1C2
	movx	a, @dptr
	acall	is_fat32_type
	jnc	found_partition
	acall	print_unknown_type

	mov	dptr, #sector_buffer + 0x1D2
	movx	a, @dptr
	acall	is_fat32_type
	jnc	found_partition
	acall	print_unknown_type

	mov	dptr, #sector_buffer + 0x1E2
	movx	a, @dptr
	acall	is_fat32_type
	jnc	found_partition
	acall	print_unknown_type

	mov	dptr, #sector_buffer + 0x1F2
	movx	a, @dptr
	acall	is_fat32_type
	jnc	found_partition
	acall	print_unknown_type
	ajmp	quit


is_fat32_type:
	add	a, #256 - 11		;is it type 0B
	jz	isf32b
	add	a, #255			;is it type 0C
	jz	isf32b
	add	a, #12
	setb	c
	ret
isf32b:
	clr	c
	ret




bad_mbr:
	acall	print_sector_buffer
	lcall	newline
	mov	dptr, #mesg_bad_mbr
	lcall	pstr
	ajmp	quit

print_unknown_type:
	push	acc
	mov	dptr, #msg_unknown_partion_type
	lcall	pstr
	pop	acc
	lcall	phex
	lcall	newline
	ret

get_byte:
	movx	a, @dptr
	inc	dptr
	ret
	
found_partition:
	inc	dptr
	inc	dptr
	inc	dptr
	inc	dptr
	acall	get_byte
	mov	partition_first_sector+0, a
	acall	get_byte
	mov	partition_first_sector+1, a
	acall	get_byte
	mov	partition_first_sector+2, a
	acall	get_byte
	mov	partition_first_sector+3, a
	acall	get_byte
	mov	partition_size+0, a
	acall	get_byte
	mov	partition_size+1, a
	acall	get_byte
	mov	partition_size+2, a
	acall	get_byte
	mov	partition_size+3, a

	mov	dptr, #mesg_partition_at
	lcall	pstr
	mov	r0, #partition_first_sector
	acall	phex32_at_r0

	lcall	newline
	ajmp	read_vol_id



bad_vol_id:
	acall	print_sector_buffer
	mov	dptr, #mesg_bad_vol_id
	lcall	pstr
	acall	print_sector_buffer
	ajmp	quit



read_vol_id:

	mov	dptr, #mesg_read_vol_id
	lcall	pstr

	mov	lba+0, partition_first_sector+0
	mov	lba+1, partition_first_sector+1
	mov	lba+2, partition_first_sector+2
	mov	lba+3, partition_first_sector+3
	
	mov	dptr, #sector_buffer
	acall	ide_read_sector

	;acall	print_sector_buffer

	mov	dptr, #sector_buffer + 0x1FE
	movx	a, @dptr
	cjne	a, #0x55, bad_vol_id
	inc	dptr
	movx	a, @dptr
	cjne	a, #0xAA, bad_vol_id
	mov	dptr, #sector_buffer + 11
	movx    a, @dptr
	jnz	bad_vol_id			;require 512 byte sector
	inc	dptr
	movx    a, @dptr
	cjne	a, #2, bad_vol_id		;require 512 byte sector
	mov	dptr, #sector_buffer + 16
	movx    a, @dptr
	cjne	a, #2, bad_vol_id		;require BPB_NumFATs == 2

	mov	dptr, #mesg_vol_id_ok
	lcall	pstr

	clr	c
	mov	dptr, #sector_buffer + 36	;offset of BPB_FATSz32
	movx	a, @dptr
	rlc	a
	mov	r2, a				;r2 to r5 has BPB_FATSz32 * 2
	inc	dptr
	movx	a, @dptr
	rlc	a
	mov	r3, a
	inc	dptr
	movx	a, @dptr
	rlc	a
	mov	r4, a
	inc	dptr
	movx	a, @dptr
	rlc	a
	mov	r5, a

	mov	dptr, #mesg_sectors_for_fats
	lcall	pstr
	mov	r0, #2
	acall	phex32_at_r0
	lcall	newline

	mov	dptr, #sector_buffer + 14	;read BPB_RsvdSecCnt
	movx	a, @dptr
	mov	reserved_sector_count+0, a
	inc	dptr
	movx	a, @dptr
	mov	reserved_sector_count+1, a

	mov	dptr, #mesg_reserved_sectors
	lcall	pstr
	mov	a, reserved_sector_count+1
	lcall	phex
	mov	a, reserved_sector_count+0
	lcall	phex
	lcall	newline

	mov	a, r2				;add (BPB_FATSz32 * 2)
	add	a, reserved_sector_count+0	;to reserved_sector_count
	mov	first_data_sector+0, a		;and store as first_data_sector
	mov	a, r3
	addc	a, reserved_sector_count+1
	mov	first_data_sector+1, a
	mov	a, r4
	addc	a, #0
	mov	first_data_sector+2, a
	mov	a, r5
	addc	a, #0
	mov	first_data_sector+3, a

	mov	a, first_data_sector+0
	add	a, partition_first_sector+0
	mov	first_data_sector+0, a
	mov	a, first_data_sector+1
	addc	a, partition_first_sector+1
	mov	first_data_sector+1, a
	mov	a, first_data_sector+2
	addc	a, partition_first_sector+2
	mov	first_data_sector+2, a
	mov	a, first_data_sector+3
	addc	a, partition_first_sector+3
	mov	first_data_sector+3, a

	mov	dptr, #mesg_first_data_sec
	lcall	pstr
	mov	r0, #first_data_sector
	acall	phex32_at_r0
	lcall	newline

	mov	dptr, #sector_buffer + 13	;read BPB_SecPerClus
	movx	a, @dptr
	mov	sectors_per_cluster, a

	mov	dptr, #mesg_sec_per_cluster
	lcall	pstr
	mov	a, sectors_per_cluster
	lcall	phex
	lcall	newline

	mov	dptr, #sector_buffer + 44	;read BPB_RootClus
	movx	a, @dptr
	mov	root_dir_cluster+0, a
	inc	dptr
	movx	a, @dptr
	mov	root_dir_cluster+1, a
	inc	dptr
	movx	a, @dptr
	mov	root_dir_cluster+2, a
	inc	dptr
	movx	a, @dptr
	mov	root_dir_cluster+3, a

	mov	dptr, #mesg_root_dir_at_cluster
	lcall	pstr
	mov	r0, #root_dir_cluster
	acall	phex32_at_r0
	lcall	newline

	ajmp	begin_player










;*************************************************************
;**                                                         **
;**                  MP3 Player Main Code                   **
;**                                                         **
;*************************************************************

begin_player:
	mov	dptr, #mesg_begin
	lcall	pstr
	clr	next_flag
	clr	prev_flag
	clr	paused_flag

	; begin with the first file in the root directory
	mov	dir_position_cluster+0, root_dir_cluster+0
	mov	dir_position_cluster+1, root_dir_cluster+1
	mov	dir_position_cluster+2, root_dir_cluster+2
	mov	dir_position_cluster+3, root_dir_cluster+3
	mov	dir_position_sector, #0
	mov	dir_position_offset, #0

traverse_dir:
	acall	fat32_get_dir_info
        mov	a, filename
	jz	stop_playing
	;acall	print_filename
	acall	is_mp3_file
	jnc	trav_dir2
	acall	play_file
trav_dir2:
	acall	trav_direction
	jnc	traverse_dir

stop_playing:
	;now what do we do?
	;let's just start over again for now
	ajmp	begin_player


trav_direction:
	jb	prev_flag, trav_prev
trav_next:
	ajmp	fat32_next_file
trav_prev:
	ajmp	fat32_prev_file






;*************************************************************
;**                                                         **
;**                       Play a File                       **
;**                                                         **
;*************************************************************


play_file:
					;print a message about the file
	mov	dptr, #mesg_playing_file
	lcall	pstr
	acall	print_filename

	clr	next_flag
	clr	prev_flag
	clr	paused_flag
					;find the first sector
	mov	r2, cluster+0
	mov	r3, cluster+1
	mov	r4, cluster+2
	mov	r5, cluster+3
	acall	fat32_cluster_to_lba

	;mov	r0, #lba
	;acall	phex32_at_r0
	;lcall	newline


	acall	bytes_to_sectors

	;mov	size+2, #0		;for testing, only play first
	;mov	size+1, #0		;64k of the file
	;mov	size+0, #128

	;now "lba" has the first sector number and "size" has the
	;number of sectors to play.  In an ideal world, we'd play
	;all the sectors within this cluster and call
	;"fat32_next_cluster" to get the next cluster's "lba", and
	;keep doing that until we've played "size" sectors.
	
	;unfortunately, we can't afford to call "fat32_next_cluster"
	;while we're playing a file, because it reads a sector and
	;that will take too much time, because we have only one
	;sector of buffering without the DRAM.

	;the approach is to just assume that all the clusters for
	;the file are in sequential order.  This means the user
	;must run defrag.  Kinda sad, because if we could buffer
	;a cluster or two (or 32 megs!) there'd be plenty of time
	;to call "fat32_next_cluster", which would also work with
	;buffered sectors in most cases.  I gotta get that DRAM
	;controller integrated into the FPGA design!



mainloop:
	mov	dptr, #sector_buffer
	acall	ide_read_sector
	acall	play_sector
	acall	inc_lba
	acall	dec_size
	jc	main_exit
	jb	next_flag, main_exit
	jb	prev_flag, main_exit
	jb	paused_flag, pause_loop

main_reenter:
	clr	paused_flag
	acall	check_buttons
	sjmp	mainloop

main_exit:
	ret


pause_loop:
	clr	playing_flag
	setb	paused_flag
	acall	check_buttons

	jb	next_flag, main_reenter
	jb	playing_flag, main_reenter
	jnb	paused_flag, main_reenter

	acall	wait
	sjmp	pause_loop



play_sector:
        mov     dptr, #dma_mp3_src
        mov	a, #sector_buffer & 255
        movx    @dptr, a                ;write LSB of src addr
        inc     dptr
        mov	a, #sector_buffer >> 8
        movx    @dptr, a                ;write MSB of src addr

        mov     dptr, #dma_mp3_count
	clr	a
        movx    @dptr, a                ;write LSB of length
        inc     dptr
        mov     a, #1
        movx    @dptr, a                ;write MSB of length

        mov     dptr, #irq_dma_mp3_ack
        movx    @dptr, a                ;clear any previous interrupt
        mov     dptr, #dma_mp3_go
        movx    @dptr, a                ;begin the transfer
        ;now we just wait for it to finish
mp3_dma_wait:
        mov     dptr, #irq_ident
        acall   wait
        movx    a, @dptr
        jnb     acc.1, mp3_dma_wait
	ret



inc_lba:
	mov	a, lba
	add	a, #1
	mov	lba, a
	mov	a, lba+1
	addc	a, #0
	mov	lba+1, a
	mov	a, lba+2
	addc	a, #0
	mov	lba+2, a
	mov	a, lba+3
	addc	a, #0
	mov	lba+3, a
	ret

	;decrement the size count, and return C=1 if zero
dec_size:
	clr	c
	mov	a, size+0
	subb	a, #1
	mov	size+0, a
	mov	a, size+1
	subb	a, #0
	mov	size+1, a
	mov	a, size+2
	subb	a, #0
	mov	size+2, a
	jnz	dec_size2
	mov	a, size+0
	jnz	dec_size2
	mov	a, size+1
	jnz	dec_size2
	setb	c
	ret
dec_size2:
	clr	c
	ret




;*************************************************************
;**                                                         **
;**                     USER Interface                      **
;**                                                         **
;*************************************************************


wait:
	jb	ri, keypress
	ret

keypress:
	lcall	cin
	lcall	upper
	lcall	cout
	mov	r2, a

check_quit:
	mov	dptr, #str_quit
	mov	r0, #state_quit
	acall	lexer
	jnc	check_next
	ajmp	quit
	ret


check_next:
	mov	dptr, #str_next
	mov	r0, #state_next
	acall	lexer
	jnc	check_prev
	acall	button_next
	ret

check_prev:
	mov	dptr, #str_previous
	mov	r0, #state_prev
	acall	lexer
	jnc	check_play
	acall	button_prev
	ret

check_play:
	mov	dptr, #str_play
	mov	r0, #state_play
	acall	lexer
	jnc	check_all_done
	acall	button_play
	ret

check_all_done:
	ret


quit:
	acall	ide_sleep
	lcall	newline
	mov	a, status
	lcall	phex
	lcall	newline
	ljmp	0x006D			;go back to PM2



;A simple fixed-string lexer.  For each character received from an
;input stream, this routine is called once for every string which
;is to be recognized within the input stream.

;  Input registers:
;    r0    Memory location to the index variable for this
;          particular string.
;    r2    The character we received.
;    dptr  Pointer to the string pattern we want to match.

;  Output registers:
;    c     Carry set if complete string found, clear if no match yet.
;    mem   Value @r0 adjusted appropriately for the next call.

lexer:
        mov     a, @r0
        inc     a               ;don't look at length byte
        movc    a, @a+dptr      ;get character from string 
        clr     c
        subb    a, r2
        jnz     lex_nope
           ;at this point, it matches the string, so now
           ;the question becomes "are we at the end?"
        inc     @r0             ;advance index
        clr     a
        movc    a, @a+dptr      ;get string length
        clr     c
        subb    a, @r0          ;subtract index
        jz      lex_match
           ;we're not at the end of the string yet
        clr     c
        ret
lex_match: ;we've found all of the string now
        mov     @r0, a          ;reset index automatically
        setb    c
        ret
lex_nope:  ;the received character doesn't match the string
           ;so we set the index back to zero, and return with
           ;carry clear
        mov     @r0, #0
        clr     c
        ret








.equ    hc165_load_pin, 0x94
.equ    hc165_clk_pin, 0x95
.equ    hc165_data_pin, 0x96



read_buttons:
	clr	hc165_load_pin
	nop
	setb	hc165_load_pin
	mov	a, #255
	mov	c, hc165_data_pin
	clr	hc165_clk_pin
	rlc	a
	setb	hc165_clk_pin
	mov	c, hc165_data_pin
	clr	hc165_clk_pin
	rlc	a
	setb	hc165_clk_pin
	mov	c, hc165_data_pin
	clr	hc165_clk_pin
	rlc	a
	setb	hc165_clk_pin
	mov	c, hc165_data_pin
	clr	hc165_clk_pin
	rlc	a
	setb	hc165_clk_pin
	mov	c, hc165_data_pin
	clr	hc165_clk_pin
	rlc	a
	setb	hc165_clk_pin
	mov	c, hc165_data_pin
	rlc	a
	cpl	a
	ret


;button 1: 02	forward
;button 2: 10	play
;button 3: 20	previous
;button 4: 08	random
;button 5: 01	vol_plus
;button 6: 04	vol_minus

check_buttons:
	acall	read_buttons
	cjne	a, button_state, ckb2
	ret
ckb2:
	mov	b, button_state
	mov	button_state, a
	jnb	acc.4, ckb3
	jnb	b.4, button_play
ckb3:
	jnb	acc.1, ckb4
	jnb	b.1, button_next
ckb4:
	jnb	acc.5, ckb5
	jnb	b.5, button_prev
ckb5:
	jnb	acc.3, ckb6
	jnb	b.3, button_rand
ckb6:
	jnb	acc.0, ckb7
	jnb	b.0, button_vol_up
ckb7:
	jnb	acc.2, ckb8
	jnb	b.2, button_vol_down
ckb8:
	ret



button_play:
	jb	playing_flag, button_pause
	mov	dptr, #mesg_play
	lcall	pstr
	setb	playing_flag
	clr	paused_flag
	ret

button_pause:
	mov	dptr, #mesg_pause
	lcall	pstr
	setb	paused_flag
	clr	playing_flag
	ret

button_next:
	mov	dptr, #mesg_next
	lcall	pstr
	setb	next_flag
	ret

button_prev:
	mov	dptr, #mesg_prev
	lcall	pstr
	setb	prev_flag
	ret

button_rand:
	mov	dptr, #mesg_rand
	lcall	pstr
	ret

button_vol_up:
	mov	dptr, #mesg_vol_up
	lcall	pstr
	ret

button_vol_down:
	mov	dptr, #mesg_vol_down
	lcall	pstr
	ret









space:
	mov	a, #' '
	ljmp	cout


	;print a 32 bit number stored @r0
phex32_at_r0:
	inc	r0
	inc	r0
	inc	r0
	mov	a, @r0
	lcall	phex
	dec	r0
	mov	a, @r0
	lcall	phex
	dec	r0
	mov	a, @r0
	lcall	phex
	dec	r0
	mov	a, @r0
	lcall	phex
	ret






;*************************************************************
;**                                                         **
;**                    FAT32 Filesystem                     **
;**                                                         **
;*************************************************************


	;lookup the next cluster of a file, given the current
	;cluster in r2 to r5.  The next cluser is returned
	;in r2 to r5.

fat32_next_cluster:

	;mov	r0, #2
	;acall	phex32_at_r0
	;mov	a, #'-'
	;lcall	cout

	mov	a, r2
	push	acc		;keep the lowest byte on the stack
	rlc	a
	mov	a, r3		;rotate bits 8 to 31 into r2 to r5
	rlc	a
	mov	r2, a
	mov	a, r4
	rlc	a
	mov	r3, a
	mov	a, r5
	anl	a, #15
	rlc	a
	mov	r4, a
	clr	a
	rlc	a
	mov	r5, a

	;mov	r0, #2
	;acall	phex32_at_r0
	;mov	a, #'-'
	;lcall	cout

				;add reserved_sector_count
	mov	a, r2
	add	a, reserved_sector_count+0
	mov	r2, a
	mov	a, r3
	addc	a, reserved_sector_count+1
	mov	r3, a
	mov	a, r4
	addc	a, #0
	mov	r4, a
	mov	a, r5
	addc	a, #0
	mov	r5, a

	;mov	r0, #2
	;acall	phex32_at_r0
	;mov	a, #'-'
	;lcall	cout
				;add partition_first_sector
	mov	a, r2
	add	a, partition_first_sector+0
	mov	r2, a
	mov	a, r3
	addc	a, partition_first_sector+1
	mov	r3, a
	mov	a, r4
	addc	a, partition_first_sector+2
	mov	r4, a
	mov	a, r5
	addc	a, partition_first_sector+3
	mov	r5, a
				;now r2 to r5 points to sector of the FAT
	mov	lba+0, r2
	mov	lba+1, r3
	mov	lba+2, r4
	mov	lba+3, r5

	;mov	r0, #2
	;acall	phex32_at_r0
	;mov	a, #'-'
	;lcall	cout
					;read the sector from the FAT
	mov	dptr, #sector_buffer
	acall	ide_read_sector

	;acall	print_sector_buffer

	pop	acc			;retrieve low byte
	anl	a, #0x7F		;only need the low 7 bits
	clr	c
	rlc	a			;mult by 4 for offset in sector
	rlc	a
	mov	r2, a
	clr	a
	rlc	a
	mov	r3, a
					;add the offset to the buffer addr
	mov	a, #sector_buffer & 255
	add	a, r2
	mov	dpl, a
	mov	a, #sector_buffer >> 8
	addc	a, r3
	mov	dph, a

	movx	a, @dptr
	mov	r2, a
	inc	dptr
	movx	a, @dptr
	mov	r3, a
	inc	dptr
	movx	a, @dptr
	mov	r4, a
	inc	dptr
	movx	a, @dptr
	anl	a, #15
	mov	r5, a

	;mov	r0, #2
	;acall	phex32_at_r0
	;lcall	newline

	ret


	;convert a cluster number in r2 to r5 to a LBA address
	;and put it into LBA.

fat32_cluster_to_lba:
	clr	c
	mov	a, r2		;first, subtract 2
	subb	a, #2
	mov	lba+0, a
	mov	a, r3
	subb	a, #0
	mov	lba+1, a
	mov	a, r4
	subb	a, #0
	mov	lba+2, a
	mov	a, r5
	anl	a, #15
	subb	a, #0
	mov	lba+3, a
				;now multiply by sectors_per_cluster
				;always power of two, so use shifts
	mov	r0, sectors_per_cluster
f32_c2l_1:
	mov	a, r0
	clr	c
	rrc	a
	mov	r0, a
	jz	f32_c2l_2
	clr	c
	mov	a, lba+0
	rlc	a
	mov	lba+0, a
	mov	a, lba+1
	rlc	a
	mov	lba+1, a
	mov	a, lba+2
	rlc	a
	mov	lba+2, a
	mov	a, lba+3
	rlc	a
	mov	lba+3, a
	sjmp	f32_c2l_1
f32_c2l_2:
	mov	a, lba+0
	add	a, first_data_sector+0
	mov	lba+0, a
	mov	a, lba+1
	addc	a, first_data_sector+1
	mov	lba+1, a
	mov	a, lba+2
	addc	a, first_data_sector+2
	mov	lba+2, a
	mov	a, lba+3
	addc	a, first_data_sector+3
	mov	lba+3, a
	ret



	;return C=0 if r2 to r5 contain an EOC marker, or C=1
	;if they do not.

fat32_is_eoc:
	clr	c
	mov	a, r2
	subb	a, #0xF8
	mov	a, r3
	subb	a, #0xFF
	mov	a, r4
	subb	a, #0xFF
	mov	a, r5
	anl	a, #15
	subb	a, #0x0F
	ret



	;advance the "dir_position_XXX" variables to the next file
	;Carry is set if end of directory, clear if not at the end

fat32_next_file:
	;more forward one position within the current sector
	mov	a, dir_position_offset
	inc	a
	anl	a, #15
	mov	dir_position_offset, a
	jnz	f32_nf_done

	;move to the next sector within the cluster
	mov	r2, sectors_per_cluster
	dec	r2
	mov	a, dir_position_sector
	inc	a
	anl	a, r2
	mov	dir_position_sector, a
	jnz	f32_nf_done

	;move to the next cluster
	mov	r2, dir_position_cluster+0
	mov	r3, dir_position_cluster+1
	mov	r4, dir_position_cluster+2
	mov	r5, dir_position_cluster+3

	acall	fat32_next_cluster

	acall	fat32_is_eoc
	jnc	f32_nf_end
	mov	dir_position_cluster+0, r2
	mov	dir_position_cluster+1, r3
	mov	dir_position_cluster+2, r4
	mov	dir_position_cluster+3, r5
f32_nf_done:
	clr	c
	ret

f32_nf_end:
	mov	dir_position_cluster+0, root_dir_cluster+0
	mov	dir_position_cluster+1, root_dir_cluster+1
	mov	dir_position_cluster+2, root_dir_cluster+2
	mov	dir_position_cluster+3, root_dir_cluster+3
	setb	c
	ret





	;back up the "dir_position_XXX" variables to the previous file
	;Carry is set if beginning of directory, clear if not at the beginning

fat32_prev_file:
	;back up one position within the current sector
	mov	a, dir_position_offset
	dec	a
	anl	a, #15
	mov	dir_position_offset, a
	orl	a, #0xF0
	cpl	a
	jnz	f32_pf_done

	;back up one sector within the current cluster
	mov	r2, sectors_per_cluster
	dec	r2
	mov	a, r2
	cpl	a
	mov	r3, a
	mov	a, dir_position_sector
	dec	a
	anl	a, r2
	mov	dir_position_sector, a
	orl	a, r3
	cpl	a
	jz	f32_pf_prev_cluster
f32_pf_done:
	clr	c
	ret



f32_pf_prev_cluster:
	;now for the really hard part... find the previous cluster!
	;FAT32 doesn't have any easy way to figure out the previous
	;cluster, and since we don't remember it (someday when the
	;DRAM controller is supported we will), the only choice is
	;to follow the cluster chain from the beginning until we get
	;to the current spot, and see what the previous cluster was.

	;"tmp_var" will remember the previous cluster, and we init
	;it with the EOC marker, in case there is no previous cluster
	mov	tmp_var+0, 0xFF
	mov	tmp_var+1, 0xFF
	mov	tmp_var+2, 0xFF
	mov	tmp_var+3, 0x0F
	mov	r2, root_dir_cluster+0
	mov	r3, root_dir_cluster+1
	mov	r4, root_dir_cluster+2
	mov	r5, root_dir_cluster+3

f32_pf_loop:
	;mov	r0, #tmp_var
	;acall	phex32_at_r0
	;acall	space
	;mov	r0, #dir_position_cluster
	;acall	phex32_at_r0
	;acall	space
	;mov	r0, #2
	;acall	phex32_at_r0
	;lcall	newline

	;if r2/r5 match "dir_position_cluster", then we're done searching
	mov	a, r2
	cjne	a, dir_position_cluster+0, f32_pf_nxt
	mov	a, r3
	cjne	a, dir_position_cluster+1, f32_pf_nxt
	mov	a, r4
	cjne	a, dir_position_cluster+2, f32_pf_nxt
	mov	a, r5
	cjne	a, dir_position_cluster+3, f32_pf_nxt
	sjmp	f32_pf_cluster
f32_pf_nxt:
	mov	tmp_var+0, r2
	mov	tmp_var+1, r3
	mov	tmp_var+2, r4
	mov	tmp_var+3, r5

	acall	fat32_next_cluster

	acall	fat32_is_eoc		;should never hit the end!
	jnc	f32_pf_error
	sjmp	f32_pf_loop


f32_pf_cluster:
	;at this point, "tmp_var" has the previous cluster.  All
	;we need to do is make sure it really is a valid cluster.
	mov	r2, tmp_var+0
	mov	r3, tmp_var+1
	mov	r4, tmp_var+2
	mov	r5, tmp_var+3
	acall	fat32_is_eoc
	jnc	f32_pf_end

	mov	dir_position_cluster+0, tmp_var+0
	mov	dir_position_cluster+1, tmp_var+1
	mov	dir_position_cluster+2, tmp_var+2
	mov	dir_position_cluster+3, tmp_var+3
	sjmp	f32_pf_done


f32_pf_error:
f32_pf_end:
	mov	dir_position_cluster+0, root_dir_cluster+0
	mov	dir_position_cluster+1, root_dir_cluster+1
	mov	dir_position_cluster+2, root_dir_cluster+2
	mov	dir_position_cluster+3, root_dir_cluster+3
	mov	dir_position_sector, #0
	mov	dir_position_offset, #0
	setb	c
	ret

















	;get the 32 bytes of info about the current file.
	;the 32 bytes are always returned at the beginning
	;of the sector buffer.

fat32_get_dir_info:
					;find the first sector of cluster
	mov	r2, dir_position_cluster+0
	mov	r3, dir_position_cluster+1
	mov	r4, dir_position_cluster+2
	mov	r5, dir_position_cluster+3
	acall	fat32_cluster_to_lba
					;add the sector offset with cluster
	mov	a, lba+0
	add	a, dir_position_sector
	mov	lba+0, a
	mov	a, lba+1
	addc	a, #0
	mov	lba+1, a
	mov	a, lba+2
	addc	a, #0
	mov	lba+2, a
	mov	a, lba+3
	addc	a, #0
	mov	lba+3, a
					;read the sector
	mov	dptr, #sector_buffer
	acall	ide_read_sector
					;set DPTR to point to dir entry
	mov	a, dir_position_offset
	anl	a, #15
	swap	a
	clr	c
	rlc	a			;mult by 32
	mov	r2, a
	clr	a
	rlc	a
	mov	r3, a
	mov	a, #sector_buffer & 255
	add	a, r2
	mov	dpl, a
	mov	a, #sector_buffer >> 8
	addc	a, r3
	mov	dph, a
					;copy the filename
	mov	r0, #filename
f32_gdi2:
	movx	a, @dptr
	inc	dptr
	mov	@r0, a
	inc	r0
	cjne	r0, #filename+11, f32_gdi2
					;copy the attrib
	movx	a, @dptr
	mov	attrib, a
					;copy first cluster number
	mov	a, dpl
	add	a, #9
	mov	dpl, a
	mov	a, dph
	addc	a, #0
	mov	dph, a
	movx	a, @dptr
	inc	dptr
	mov	cluster+2, a
	movx	a, @dptr
	inc	dptr
	mov	cluster+3, a
	inc	dptr			;skip write time and date
	inc	dptr
	inc	dptr
	inc	dptr
	movx	a, @dptr
	inc	dptr
	mov	cluster+0, a
	movx	a, @dptr
	inc	dptr
	mov	cluster+1, a
					;copy file size
	movx	a, @dptr
	inc	dptr
	mov	size+0, a
	movx	a, @dptr
	inc	dptr
	mov	size+1, a
	movx	a, @dptr
	inc	dptr
	mov	size+2, a
	movx	a, @dptr
	inc	dptr
	mov	size+3, a
	ret




	;return C=1 if it's a MP3 file, or C=0 if not
is_mp3_file:
	mov	a, filename+0
	jz	isfile_no
	add	a, #256 - 0xE5
	jz	isfile_no
	mov	a, attrib
	jb	acc.1, isfile_no	;ATTR_HIDDEN
	jb	acc.2, isfile_no	;ATTR_SYSTEM
	jb	acc.3, isfile_no	;ATTR_VOLUME_ID
	jb	acc.4, isfile_no	;ATTR_DIRECTORY
	mov	a, filename+8
	cjne	a, #'M', isfile_no
	mov	a, filename+9
	cjne	a, #'P', isfile_no
	mov	a, filename+10
	cjne	a, #'3', isfile_no
	setb	c
	ret
isfile_no:
	clr	c
	ret


	;print the filename, attrib, first cluster, and file size

print_filename:
	mov	r0, #filename
pfn1:
	mov	a, @r0
	lcall	cout
	inc	r0
	cjne	r0, #filename+8, pfn1
	mov	a, #'.'
	lcall	cout
pfn2:
	mov	a, @r0
	lcall	cout
	inc	r0
	cjne	r0, #filename+11, pfn2

	mov	dptr, #mesg_attrib
	lcall	pstr
	mov	a, attrib
	lcall	phex
	mov	dptr, #mesg_cluster
	lcall	pstr
	mov	r0, #cluster
	acall	phex32_at_r0
	mov	dptr, #mesg_size
	lcall	pstr
	mov	r0, #size
	acall	phex32_at_r0
	lcall	newline

	mov	r0, #dir_position_cluster
	acall	phex32_at_r0
	acall	space
	mov	a, dir_position_sector
	lcall	phex
	acall	space
	mov	a, dir_position_offset
	lcall	phex
	lcall	newline

	ret


	;convert "size" in bytes to the number of sectors.  It's
	;just a divide by 512 with rounding up

bytes_to_sectors:
	mov	r2, size+0	;save lower 9 bits
	mov	a, size+1
	anl	a, #1
	mov	r3, a

	clr	c		;rotate right 1 bit
	mov	a, size+3
	rrc	a
	mov	size+3, a
	mov	a, size+2
	rrc	a
	mov	size+2, a
	mov	a, size+1
	rrc	a
	mov	size+1, a
	mov	size+0, size+1	;rotate another 8 bits
	mov	size+1, size+2
	mov	size+2, size+3
	mov	size+3, #0
	mov	a, r2
	jnz	b2s_inc
	mov	a, r3
	jnz	b2s_inc
	ret
b2s_inc:			;round up
	mov	a, size+0
	add	a, #1
	mov	size+0, a
	mov	a, size+1
	addc	a, #0
	mov	size+1, a
	mov	a, size+2
	addc	a, #0
	mov	size+2, a
	ret










;*************************************************************
;**                                                         **
;**              IDE Disk Drive Driver Code                 **
;**                                                         **
;*************************************************************


ide_read_4k_to_8000:







	;read a sector from the drive, and put it @dptr
ide_read_sector:
        push    dph
        push    dpl
        acall   ide_busy                ;make sure drive is ready
        acall   wr_lba                  ;tell it which sector we want
        mov     dptr, #ide_command
        mov     a, #ide_cmd_read
        movx    @dptr, a                ;ask the drive to read it
        acall   ide_drq                 ;wait until it's got the data
        jb      acc.0, get_err
        pop     dpl
        pop     dph
        acall   read_data               ;grab the data
        clr     a
        ret


        ;when an error occurs, we get acc.0 set from a call to ide_drq
        ;or ide_busy (which read the drive's status register).  If
        ;that error bit is set, we should jump here to read the drive's
        ;explaination of the error, to be returned to the user.  If for
        ;some reason the error code is zero (shouldn't happen), we'll
        ;return 255, so that the main program can always depend on a
        ;return of zero to indicate success.
get_err:mov     dptr, #ide_err
        movx    a, @dptr
        jz      gerr2
        ret
gerr2:  mov     a, #255
        ret



	;ask the drive to run in it's lowest current mode
	;does this work?  Need to fiddle with it some more.

ide_set_low_power:
	;acall	ide_drq
	;mov	dptr, #ide_features
	;mov	a, #0x9A		;set current subcommand
	;movx	@dptr, a
	;mov	dptr, #ide_sec_cnt
	;mov	a, #1			;lowest available current
	;movx	@dptr, a
	;mov	dptr, #ide_command
	;mov	a, #ide_cmd_set_features
	;movx	@dptr, a
	ret

	;put the drive into sleep mode

ide_sleep:
	; acall	ide_busy
	mov	dptr, #ide_command
	mov	a, #ide_cmd_sleep
	movx	@dptr, a
	;acall	ide_busy
	mov	dptr, #ide_status
	movx	a, @dptr
	ret


ide_init:
ideiwait:
	mov	dptr, #ide_head
	mov	a, #10100000b
	movx	@dptr, a		;select the master device
	mov	dptr, #ide_status
	movx	a, @dptr
	  push	acc
	  mov	a, #'.'			;should slow down the printing of dots
	  lcall	cout
	  acall	wait
	  pop	acc
	jnb	acc.6, ideiwait         ;wait for RDY bit to be set
	jb      acc.7, ideiwait         ;wait for BSY bit to be clear
	ret


        ;do a hard reset on the drive, by pulsing its reset pin.
        ;this should usually be followed with a call to "ide_init".
	;it's not included inside ide_init because it may be
	;desirable to call this early in the boot up sequence, and
	;then initialize other hardware or memory while the drives
	;start spinning up their motors.
ide_hard_reset:
        mov     dptr, #ide_rst_bit
        clr     a
        movx    @dptr, a                ;assert the reset pin
        acall   delay_10ms              ;IDE requires reset > 25 ms
        acall   delay_10ms
        acall   delay_10ms
        mov     a, #1
        movx    @dptr, a                ;de-assert the reset pin
        acall   delay_10ms              ;brief delay while drive inits
        ret

        ;delay for 10 ms
        ;10ms at 7.3728 MHz is 6144 machine cycles
        ;6144 = 256 * 24
delay_10ms:
        mov     r0, #0
d1ms:   mov     r1, #10         ;1 cycle
        djnz    r1, *           ;20 cycles
        nop                     ;1 cycle
        djnz    r0, d1ms        ;2 cycles
five_cy:nop
        ret






ide_busy:
	mov	dptr, #ide_status
	movx	a, @dptr
	  push	acc
	  ;mov	a, #'.'
	  ;lcall	cout
	  acall	wait
	  pop	acc
	jb	acc.7, ide_busy		;wait for BSY to be clear
	ret


ide_drq:
	mov	dptr, #ide_status
	movx	a, @dptr
	  push	acc
	  ;mov	a, #'_'
	  ;lcall	cout
	  acall	wait
	  pop	acc
	jb	acc.7, ide_drq		;wait for BSY to be clear
	jnb     acc.3, ide_drq          ;wait for DRQ to be set
	ret



        ;write the logical block address to the drive's registers
wr_lba:
        mov     a, lba+3
        anl     a, #0x0F
        orl     a, #0xE0
        mov     dptr, #ide_head
        movx    @dptr, a
        mov     a, lba+2
        mov     dptr, #ide_cyl_msb
        movx    @dptr, a
        mov     a, lba+1
        mov     dptr, #ide_cyl_lsb
        movx    @dptr, a
        mov     a, lba+0
        mov     dptr, #ide_sector
        movx    @dptr, a
        mov     a, #1
        mov     dptr, #ide_sec_cnt
        movx    @dptr, a
        ret





print_sector_buffer:
	mov	r2, #32
	mov	dptr, #sector_buffer


;print a hexdump of memory.  R2 specifies how many 16 byte
;lines to print.  DPTR specifies where to dump from.

hexdump:
	lcall	newline
hd1:	lcall	phex16			;tell 'em the memory location
	mov	a, #':'
	lcall	cout
	mov	a, #' '
	lcall	cout
	mov	r3, #16			;r3 counts # of bytes to print
hd2:	movx	a, @dptr
	inc	dptr
	lcall	phex
	lcall	space
	djnz	r3, hd2
	clr	c
	mov	a, dpl
	subb	a, #16
	mov	dpl, a
	mov	a, dph
	subb	a, #0
	mov	dph, a

	mov	r3, #16
hd3:	movx	a, @dptr
	inc	dptr
	anl	a, #0x7F
	cjne	a, #0x7F, hd3b
	clr	a
hd3b:	add	a, #224
	jc	hd4
	clr	a
hd4:	add	a, #32
	lcall	cout
	djnz	r3, hd3
	lcall	newline
	djnz	r2, hd1
	ret




        ;Read a block of 512 bytes (one sector) from the drive
        ;and store it in memory @ DPTR
read_data:
        ;ajmp   read_data_pio


        ;read 512 bytes of data using fast DMA
read_data_dma:
        mov     a, dpl
        mov     b, dph
        mov     dptr, #dma_ide_dest
        movx    @dptr, a                ;write LSB of dest addr
        inc     dptr
        mov     a, b
        movx    @dptr, a                ;write LSB of dest addr
        mov     dptr, #dma_ide_count
        clr     a
        movx    @dptr, a
        inc     dptr
        mov     a, #1
        movx    @dptr, a                ;request 256 16 bit reads
        mov     dptr, #irq_dma_ide_ack
        movx    @dptr, a                ;clear any previous interrupt
        mov     dptr, #dma_ide_go
        movx    @dptr, a                ;begin the transfer
        ;now we just wait for it to finish
        mov     dptr, #irq_ident
dma_wait:
        movx    a, @dptr                ;2  1 psen = 2 bytes transfered
        jnb     acc.0, dma_wait         ;2  3 psen = 6 bytes transfered
        ret

        ;speed = 8 bytes per 4 cycles -> 1229 kbyte/sec
        ;  remember, this is raw transfer speed, and doesn't
        ;  include the time to set up the transfer.

        ;Ideally, we'd be executing real code instead of this
        ;busy loop.  With some luck, real code would have more
        ;psen's per second than movx (only mul and div have less).
        ;If we call into PM2, the psens (and dma) will stall!

        ;Maybe someday the DMA controller will be able to detect
        ;unused bus time and use it... that'd double the speed for
        ;this test case.  Will need to investigate actual speed once
        ;the drivers are written and somewhat mature.  A really
        ;ambitious approach would be to watch the psen fetch and
        ;do a little table lookup of the opcode, and let the state
        ;machine run full speed for the entire instruction and
        ;into the fetch of the next opcode, so we could use all
        ;the available time.  Maybe someday when I have lots of
        ;free time (highly unlikely).



        ;read 512 bytes of data using slow I/O
read_data_pio:
        mov     r5, #0
rdblk2: push    dph                     ;2
        push    dpl                     ;2
        mov     dptr, #ide_data         ;2
        movx    a, @dptr                ;2  read lsb
        mov     r2, a                   ;1
        mov     dptr, #ide_data_buf_msb ;2
        movx    a, @dptr                ;2  read msb
        mov     r3, a                   ;1
        pop     dpl                     ;2
        pop     dph                     ;2
        mov     a, r2                   ;1
        movx    @dptr, a                ;2  write lsb
        inc     dptr                    ;2
        mov     a, r3                   ;1
        movx    @dptr, a                ;2  write msb
        inc     dptr                    ;2
        djnz    r5, rdblk2              ;2
        ret

        ;speed = 2 bytes per 30 cycles -> 41 kbyte/sec





















;*************************************************************
;**                                                         **
;**                     Memory Manager                      **
;**                                                         **
;*************************************************************


        ;initialize the memory manager, figure out what
        ;SIMM is installed, and initialize the memory
        ;block descriptor table.  The first 32 blocks
        ;are always used for the table itself, regardless
        ;of what SIMM is installed.  All the rest are
        ;either free blocks, or not implemented by the SIMM.
        ;The xilinx chip needs to be configured with the
        ;DRAM controller before calling here.
        ;return: C=1 if an error (no SIMM), C=0 for success

init_memory_mgr:
	clr	a
	mov	r2, a
	mov	r3, a
        acall   map_block
        acall   write_test_bytes
        acall   check_test_bytes
        jnc     initmm2
        mov     dptr, #msg_bad_simm
        lcall   pstr
        ret                     ;error, no simm is installed
initmm2:
        ;in a smaller 4 or 8 meg simm, writing to block 2048 will
        ;overwrite whatever's in block 0, but 16 and 32 meg simms
        ;have a block 2048, so writing to block 0, then 2048, and
        ;checking to see if block 0 still is correct will tell if
        ;the simm is (4 or 8) vs (16 or 32).
	mov	r2, #2048 & 255
	mov	r3, #2048 >> 8
	clr	a
        acall   map_block
        acall   write_test_bytes
	clr	a
	mov	r2, a
	mov	r3, a
        acall   map_block
        acall   check_test_bytes
        mov     acc.1, c
        push    acc

        ;double sided simms (8 and 32 megs) implement block 1024
        ;whereas single sided simms can't implement block 1024, so
        ;writing and checking 1024 will tell if it's a single
        ;(4 or 16) or double (8 or 32).
	mov	r2, #1024 & 255
	mov	r3, #1024 >> 8
	clr	a
        acall   map_block
        acall   write_test_bytes
        acall   check_test_bytes
        pop     acc
        mov     acc.0, c
        cpl     a
        anl     a, #3
        mov     r4, a           ;r4 has simm size, 0 to 3
        add     a, #2
        movc    a, @a+pc        ;now Acc has simm size, in Mbytes
        sjmp    initmm3
        .db     4, 8, 16, 32
initmm3:
        ;print a message here about which simm we found
        push    acc
        mov     dptr, #msg_found
        lcall   pstr
        pop     acc
        lcall   pint8
        mov     dptr, #msg_size
        lcall   pstr


        mov     r5, #0          ;r5 will count which 1MB section (0-31)
initmm4:
        mov     a, r5
        acall   is_implemented
        jz      initmm9         ;don't bother if not implemented

        mov     a, r5
	mov	r2, a
	clr	a
	mov	r3, a
        acall   map_block

        ;mov     a, #'p'
        ;acall   cout
        ;mov     a, #'='
        ;acall   cout
        ;mov     a, r5
        ;lcall   phex

initmm5:
        ;whatever page is mapped at 1000 to 1FFF is the previous
        ;page where we didn't write the last entry, but now we
        ;know what it will be, so write it now.
        mov     a, r5
        jz      initmm6         ;don't do this the first time
        mov     dptr, #0x1FF0
        clr     a
        movx    @dptr, a
        inc     dptr
        mov     a, r5
        movx    @dptr, a
initmm6:
        ;first, let's just rapidly fill up the linked list
        ;knowing that the last one may be wrong, since we
        ;don't know where it will point yet.  We do know
        ;that each block will always point to the next in
        ;a 1 meg page, at least initially when all the
        ;memory is free and we're just initializing.
        mov     r1, #0          ;r1 counts page within that 1MB (0-255)
        mov     dptr, #0
initmm7:
        inc     r1
        mov     a, r1
        movx    @dptr, a
        inc     dptr
        mov     a, r5
        movx    @dptr, a
        mov     a, dpl
        add     a, #15
        mov     dpl, a
        mov     a, dph
        addc    a, #0
        mov     dph, a
        anl     a, #0xF0
        jz      initmm7

        ;now map this page at 1000 to 1FFF, so later we can write
        ;the last entry to point to the next 1MB part
	mov	a, r5
	mov	r2, a
	mov	r3, #0
	mov	a, #1
        acall   map_block
initmm9:
        inc     r5
        cjne    r5, #32, initmm4

        ;the very last entry needs a NULL to terminate the list
        mov     dptr, #0x1FF0
        clr     a
        movx    @dptr, a
        inc     dptr
        movx    @dptr, a

        lcall   newline

        ;now we need to make the first 32 pages not free,
        ;since they're used for tracking what memory is
        ;used and what's free
	clr	a
	mov	r2, a
	mov	r3, a
        acall   map_block
        mov     dptr, #0x01F0
        clr     a
        movx    @dptr, a
        inc     dptr
        movx    @dptr, a

        ;I hope that worked!
        ret




        ;This function returns zero if a particular simm
        ;doesn't implement a particlar 1 MB memory range,
        ;or non-zero if this simm implements that area.
        ;inputs:
        ;       acc = 1 MB memory area (0-32)
        ;       r4 = simm type (0=4M, 1=8M, 2=16M, 3=32M)
is_implemented:
        rr      a
        rr      a
        anl     a, #7
        mov     r0, a
        mov     a, r4
        anl     a, #3
        add     a, #2
        movc    a, @a+pc
        sjmp    is_imp2
        .db     00000001b       ;4 meg = blocks 0 to 1023
        .db     00000011b       ;8 meg = blocks 0 to 2047
        .db     01010101b       ;16 meg = blocks 0 to 1023
                                ;       blocks 2048 to 3071
                                ;       blocks 4096 to 5119
                                ;       blocks 6144 to 7167
        .db     11111111b       ;32 meg = blocks 0 to 8191
is_imp2:
        cjne    r0, #0, is_imp3
        anl     a, #1
        ret
is_imp3:rr      a
        dec     r0
        sjmp    is_imp2




        ;write a simple test to a DRAM page, just write
        ;the page number a couple times, with different
        ;XOR'd with different bit patterns.  This way,
        ;we are unlikely to read garbage as good, and
        ;it will always detect a page that is corrupted
        ;by writing to a different page (small simms
        ;which appear repeated 4 times)
write_test_bytes:
        mov     dptr, #dram_page_cfg
        movx    a, @dptr        ;read page # lsb
        mov     b, a
        inc     dptr
        movx    a, @dptr        ;read page # msb
        mov     dptr, #0
        xrl     a, #0x5A
        movx    @dptr, a
        inc     dptr
        cpl     a
        movx    @dptr, a
        inc     dptr
        mov     a, b
        xrl     a, #0x9C
        movx    @dptr, a
        inc     dptr
        cpl     a
        movx    @dptr, a
        ret

        ;read a page previously written with "write_test_bytes"
        ;and return C=0 if good, C=1 if wrong bytes read back
check_test_bytes:
        mov     dptr, #dram_page_cfg
        movx    a, @dptr        ;read page # lsb
        mov     r0, a
        inc     dptr
        movx    a, @dptr        ;read page # msb
        mov     b, a
        mov     dptr, #0
        movx    a, @dptr
        xrl     a, #0x5A
        cjne    a, b, cktb_err
        inc     dptr
        movx    a, @dptr
        xrl     a, #0xA5
        cjne    a, b, cktb_err
        inc     dptr
        mov     b, r0
        movx    a, @dptr
        xrl     a, #0x9C
        cjne    a, b, cktb_err
        inc     dptr
        movx    a, @dptr
        xrl     a, #0x63
        cjne    a, b, cktb_err
        clr     c
        ret
cktb_err:
        setb    c
        ret






map_block:
        anl     a, #15
        rl      a
        add     a, #dram_page_cfg & 255
        mov     dpl, a
        clr     a
        addc    a, #dram_page_cfg >> 8
        mov     dph, a
        mov     a, r2
        movx    @dptr, a
        inc     dptr
        mov     a, r3
        movx    @dptr, a
        ret







;*************************************************************
;**                                                         **
;**                 String and Constants                    **
;**                                                         **
;*************************************************************




mesg_read_mbr:
	.db	"Reading Master Boot Record",13,10,0
mesg_bad_mbr:
	.db	"MBR Signature Not Found",13,10,0
msg_unknown_partion_type:
	.db	"Unknown Partition Type: ",0
mesg_partition_at:
	.db	"Found FAT32 Partition at LBA = ",0

mesg_read_vol_id:
	.db	"Reading FAT32 Volume ID",13,10,0
mesg_bad_vol_id:
	.db	"Bad FAT32 Volume ID",13,10,0
mesg_first_data_sec:
	.db	"First Data Sector at LBA = ",0

mesg_playing_file:
	.db	"Playing file: ",0

mesg_vol_id_ok:
	.db	"Volume ID looks ok",13,10,0
mesg_sec_per_cluster:
	.db	"Sectors per cluster: ",0
mesg_root_dir_at_cluster:
	.db	"Root directory starts at cluster # ",0
mesg_sectors_for_fats:
	.db	"Sectors used for FAT tables: ",0
mesg_reserved_sectors:
	.db	"Reserved Sectors: ",0
mesg_attrib:
	.db	"  Attrib=",0
mesg_cluster:
	.db	"  First_Cluster=",0
mesg_size:
	.db	"  Size=",0
mesg_welcome:
	.db	13,10,13,10,13,10
	.db	"MP3 Player, Version 0.5.0, Dec 14, 2000",13,10
	.db	"Copyright (C) 2000, PJRC.COM, LLC",10,13,10,13
	.db	"Type 'QUIT' to return to the monitor",13,10,13,10,0
mesg_begin:
	.db	13,10,"MP3 Player, Begin Playback",13,10,0
mesg_xilinx_cfg:
	.db	"Configuring Xilinx XCS10XL FPGA Chip: ",0
mesg_ok:
	.db	"Ok",13,10,0
mesg_error:
	.db	"Error!",13,10,0
mesg_reset_ide:
	.db	"Reseting IDE Devices",13,10,0
mesg_sta013_cfg:
	.db	"Configuring STA013 MP3 Decoder Chip: ",0
mesg_ide_init:
	.db	"Waiting for Hard Drive to be ready",13,10,0

mesg_sector_buffer:
	.db	13,10,"Current Sector Buffer Contents:",13,10,0

mesg_play:
	.db	"Play",13,10,0
mesg_pause:
	.db	"Pause",13,10,0
mesg_next:
	.db	"Next File",13,10,0
mesg_prev:
	.db	"Previous File",13,10,0
mesg_rand:
	.db	"Random",13,10,0
mesg_vol_up:
	.db	"Vol Up",13,10,0
mesg_vol_down:
	.db	"Vol Down",13,10,0


str_quit:
	.db	4, "QUIT"
str_play:
	.db	4, "PLAY"
str_next:
	.db	4, "NEXT"
str_previous:
	.db	8, "PREVIOUS"

mesg_init_mm:
	.db	"Initialize Memory Manager",13,10,0
msg_bad_simm:
        .db     "Bad or Missing SIMM",13,10,0
msg_found:
        .db     "Found ",0
msg_size:
        .db     " meg SIMM",13,10,0




.org    location+0xF00
.db     0xA5,0xE5,0xE0,0xA5     ;signiture bytes
.db     253,0,0,0               ;id
.db     0,0,0,0                 ;reserved
.db     0,0,0,0                 ;reserved
.db     0,0,0,0                 ;reserved
.db     0,0,0,0                 ;reserved
.db     0,0,0,0                 ;user defined
.db     255,255,255,255         ;length and checksum (255=unused)
.db     "Auto-Start",0
.org    location+0xF00+64       ;executable code begins here

	ljmp	begin
	ret


